@@ -2,7 +2,7 @@ package DBIx::TextIndex::QueryParser;
 
 use strict;
 
-our $VERSION = '0.19';
+our $VERSION = '0.20';
 
 use DBIx::TextIndex::Exception;
 use Text::Balanced qw(extract_bracketed extract_delimited);
@@ -16,68 +16,91 @@ sub new {
     return $self;
 }
 
+sub term_fields {
+    my $self = shift;
+    return sort { $a cmp $b } keys %{$self->{TERM_FIELDS}};
+}
+
 sub parse {
     my $self = shift;
-    my $text = shift;
+    delete($self->{TERM_FIELDS});
+    $self->_parse(@_);
+}
+
+sub _parse {
+    my $self = shift;
+    my $q = shift;
     my @clauses;
 
-    $text =~ s/\s+$//;
+    $q =~ s/\s+$//;
 
-    while ($text) {
+    while ($q) {
 	my $clause;
 
-	if ($text =~ s/^\s+//) {
+	if ($q =~ s/^\s+//) {
 	    next;
 	}
 
-	if ($text =~ s/^(AND|OR)\s+//) {
+	if ($q =~ s/^(AND|OR)\s+//) {
 	    $clause->{CONJ} = $1;
 	}
 
-	if ($text =~ s/^\+//) {
+	if ($q =~ s/^\+//) {
 	    $clause->{MODIFIER} = 'AND';
-	} elsif ($text =~ s/^\-//) {
+	} elsif ($q =~ s/^\-//) {
 	    $clause->{MODIFIER} = 'NOT';
 	} else {
 	    $clause->{MODIFIER} = 'OR';
 	}
 
-	if ($text =~ m/^\(/) {
-	    my ($extract, $remain) = extract_bracketed($text, "(");
+	if ($q =~ s/^(\w+)://) {
+	    $clause->{FIELD} = $1;
+	    $self->{TERM_FIELDS}->{$clause->{FIELD}}++;
+	} else {
+	    $self->{TERM_FIELDS}->{__DEFAULT}++;
+	}
+
+	if ($q =~ m/^\(/) {
+	    my ($extract, $remain) = extract_bracketed($q, "(");
 	    unless ($extract) {
 		# FIXME: hard coded error message
 		throw $QRY( error => 'Open and close parentheses are uneven.');
 	    }
-	    $text = $remain;
+	    $q = $remain;
 	    $extract =~ s/^\(//;
 	    $extract =~ s/\)$//;
 	    $clause->{TYPE} = 'QUERY';
-	    $clause->{QUERY} = $self->parse($extract);
-	} elsif ($text =~ m/^\"/) {
-	    my ($extract, $remain) = extract_delimited($text, '"');
+	    $clause->{QUERY} = $self->_parse($extract);
+	} elsif ($q =~ m/^\"/) {
+	    my ($extract, $remain) = extract_delimited($q, '"');
 	    unless ($extract) {
 		# FIXME: hard coded error message
 		throw $QRY( error => 'Quotes must be used in matching pairs.')
 	    }
-	    $text = $remain;
+	    $q = $remain;
 	    $extract =~ s/^\"//;
 	    $extract =~ s/\"$//;
 	    $clause->{TYPE} = 'PHRASE';
 	    $clause->{TERM} = $extract;
-	    $clause->{PHRASETERMS} = $self->parse($extract);
-	} elsif ($text =~ s/^(\S+[\-\&\.\@\'\*]\S+)//) {
+	    $clause->{PHRASETERMS} = $self->_parse($extract);
+	    if ($q =~ s/^~(\d+)//) {
+		$clause->{PROXIMITY} = $1;
+	    } else {
+		$clause->{PROXIMITY} = 1;
+	    }
+	} elsif ($q =~ s/^(\S+[\-\&\.\@\'\*]\S+)//) {
 	    $clause->{TYPE} = 'IMPLICITPHRASE';
 	    $clause->{TERM} = $1;
 	    $clause->{PHRASETERMS} =
-	     $self->parse(join(' ', split('[\-\&\.\@\'\*]', $clause->{TERM})));
-	} elsif ($text =~ s/^(\S+)\?//) {
+	     $self->_parse(join(' ', split('[\-\&\.\@\'\*]',$clause->{TERM})));
+	} elsif ($q =~ s/^(\S+)\?//) {
 	    $clause->{TYPE} = 'PLURAL';
 	    $clause->{TERM} = $1;
-	} elsif ($text =~ s/^(\S+)\*\s*//) {
+	} elsif ($q =~ s/^(\S+)\*\s*//) {
 	    $clause->{TYPE} = 'WILD';
 	    $clause->{TERM} = $1;
 	} else {
-	    $text =~ s/(\S+)//;
+	    $q =~ s/(\S+)//;
 	    $clause->{TYPE} = 'TERM';
 	    $clause->{TERM} = $1;
 	}
